/* Heuristica das Economias
 * Este programa encontra uma solucao para o problema do caixeiro viajante utilizando o metodo das economias.
 * O metodo consiste em agrupar cidades sucessivamente, de forma sequencial, para formar um roteiro, com base em uma ordem
 * decrescente do calculo de economias decorrentes da sua uniao, que considera a distancia de cada um dos pontos ao ponto
 * inicial e tambem a distancia entre eles.
 * Autora: Larissa Alves Petri     #USP: 5117594
 * e-mail: lariss@gmail.com
 * Data: /06/2006
 */


// Bibliotecas utilizadas:

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>


// Definicao das constantes:

#define N 6 // Numero de cidades do problema.
#define T 30 // Tamanho do nome do arquivo de entrada.
#define PI 3.141592 // Valor de pi.
#define R 6378.388 
#define INF 32000000 // Infinito.


// Declaracao da estrutura:

typedef struct economia
/* Estrutura da lista de economias.
 * i: cidade onde se inicia a aresta.
 * f: cidade onde termina a aresta.
 * s: valor da economia se a cidade i for ligada diretamente a cidade f, sem pasar pela cidade inicial.
 */
{
	int i, f, s;
} economia; // Variavel do tipo economia.


// Declaracao das funcoes:

void le_arq (float x[N], float y[N], char c[N])
/* Funcao que le o arquivo de entrada, que  contem as coordenadas das cidades. Esse arquivo deve ser .txt.
 * x[N] e y[N]: vetores que receberao as coordenadas.
 * c[N]: nome do arquivo.
 */
{
	int i;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as coordenadas e guarda-as em dois vetores, x e y.
	for (i = 0; i < N; i++)
		fscanf (e, "%*d %f %f\n", &x[i], &y[i]);

	// Fecha o arquivo.
	fclose (e);
}

void le_arq_matriz (int **d, char c[N])
/* Funcao que le o arquivo de entrada, que  contem as distancias entra as cidades. Esse arquivo deve ser .txt.
 * **d: matriz que recebera as distancias.
 * c[N]: nome do arquivo.
 */
{
	int i, j;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as distancias e guarda-as na matriz d.
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < (N - 1); j++)
			fscanf (e, " %d", &d[i][j]);
		fscanf (e, " %d\n", &d[i][j]);
	}

	// Fecha o arquivo.
	fclose (e);
}

void att (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia pseudo-euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo ATT).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, t[N][N];
	float xd, yd, r[N][N];

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			r[i][j] = sqrt ((xd * xd + yd * yd) / 10.0);
			t[i][j] = (int)(r[i][j]);
			if (t[i][j] < r[i][j])
				d[i][j] = t[i][j] + 1;
			else
				d[i][j] = t[i][j];
		}
}

void euc (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo EUC_2D).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j;
	float xd, yd;
	
	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			d[i][j] = (int)(sqrt (xd * xd + yd * yd));
		}
}

void geo (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia geometrica entre as cidades e guarda-a na matriz d (arquivos do tipo GEO).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, deg;
	float min, q1, q2, q3;

	for (i = 0; i < N; i++)
	{
		deg = (int)(x[i]);
		min = x[i] - deg;
		x[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
		deg = (int)(y[i]);
		min = y[i] - deg;
		y[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
	}

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			q1 = cos (y[i] - y[j]);
			q2 = cos (x[i] - x[j]);
			q3 = cos (x[i] + x[j]);
			d[i][j] = (int)(R * acos (0.5 * ((1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0);
		}
}

int verifica (int v[N], int e)
/* Funcao que retorna a posicao do elemento e no vetor v.
 * v[N]: vetor solucao.
 * e: elemento do vetor.
 */
{
	int i = 0;
	
	// Percorre o vetor, procurando pelo elemento.
	while ((i < N) && (v[i] != e))
		i++;
	
	// Retorna a posicao do elemento.
	return i;
}

long int posicao (economia *e, int m, long int n)
/* Funcao que retorna a posicao em que m deve ser inserido na lista de economia.
 * *e: ponteiro para a lista de economias.
 * m: elemento que sera inserido na lista.
 * n: tamanho da lista de economias.
 */
{
	long int i, j;
    
	// Inicializacao das variaveis.
	j = (-1) * INF;
	i = 0;
	
	// Percorre a lista ate encontrar a posicao em que o elemento deve ser inserido.
	while ((i < (n-1)) && (e[i].s != j))
    	{
        	if (m >= e[i].s)
			break;
	        i++;
    	}
	
	// Retorna a posicao.
	return i;
}

void insere_vetor (int v[N], int e, int p)
/* Funcao que insere o elemento e na posicao p do vetor v.
 * v[N]: vetor solucao.
 * e: elemento a ser inserido no vetor.
 * p: posicao em que deve ser inserido o elemento.
 */
{
	int i, j, k;

	// Inicializacao das variaveis.
	j = v[p];
	i = p + 1;
    
	// Rearranjo dos elementos do vetor, para insercao do novo elemento no meio do vetor.
	while (i < N)
	{
    		k = v[i];
	        v[i] = j;
        	j = k;
		// Testa se ainda ha elementos, nao nulos, no vetor.
        	if (!j)
		        i = N;
		        i++;
    	}

	v[p] = e;
}

void insere_lista (economia *e, int s, int i, int f, int p, long int n)
/* Funcao que insere o elemento e na posicao p da lista de economia.
 * *e: ponteiro para a lista de economias.
 * s: valor da economia.
 * i: cidade inicial.
 * f: cidade final.
 * p: posicao em que os elementos serao inseridos.
 * n: tamanho da lista.
 */
{
	int js, ji, jf, ks, ki, kf;
	long int m, q;

	// Inicializacao das variaveis.
	q = (-1) * INF;
	js = e[p].s;
	ji = e[p].i;
	jf = e[p].f;
	m = p + 1;
	
	// Rearranjo dos elementos da lista, para insercao dos novos elementos no meio da lista.
	while (m < n)
    	{
    		ks = e[m].s;
	    	ki = e[m].i;
    		kf = e[m].f;
	        e[m].s = js;
        	e[m].i = ji;
	        e[m].f = jf;
        	js = ks;
	        ji = ki;
        	jf = kf;
	        if (js == q)
        		m = n;
	        m++;
    	}
	
	// Insercao dos novos elemento.
	e[p].s = s;
	e[p].i = i + 1;
	e[p].f = f + 1;
}

void econ (int v[N], int **d, int r)
/* Funcao que gera uma solucao inicial pela heuristica das economias.
 * v[N]: vetor que recebera a solucao inicial.
 * **d: matriz que contem a distancia entre as cidades.
 * r: cidade em que se inicia o ciclo.
 */
{
	int i, j, k, m, p, ma[N][N];
	long int n, q;
    	economia *e;
		
	// Calculo do tamanho da lista de economias.
	n = 0;
	for (i = 1; i <= (N - 2); i++)
        	n = n + i;
	
	// Alocacao de memoria para a lista de economias.
	e = (economia *) malloc (n * sizeof (economia));
	if (!e)
	{
        	printf ("Memoria Insuficiente!\n"); // Mensagem de erro.
        	exit (0);
    	}

	// Inicializacao da lista de economias, com -infinito.
	for (i = 0; i < n; i++)
        	e[i].s = (-1) * INF;
	
	// Percorre o vetor.
  	for (i = 0; i < (N - 1); i++)	
		// Verifica se a cidade i nao eh a cidade inicial.
		if (i != (r - 1))
	  	   	for (j = (i + 1); j < N; j++)
				// Verifica se a cidade j nao eh a cidade inicial.
				if (j != (r - 1))
	  		   	{
					// Calculo do valor da economia, se a cidade i for ligada diretamente a cidade j,
					// sem pasar pela cidade inicial.
        	   			m = d[0][i] + d[0][j] - d[i][j];
					// Posicao em que a economia deve ser inserida na lista.
           				q = posicao (e, m, n);
					// Insercao do elemento m na lista de economias, na posicao p.
		      			insere_lista (e, m, i, j, q, n);
	    			}
 
	// Inicializacao da matriz de adjacencias.
	for (i = 0; i < N; i++)
        	for (j = i; j < N; j++)
            		ma[i][j] = ma [j][i] = 0;
        
	// Inicializacao do vetor solucao inicial.
	v[0] = r;
	v[1] = e[0].i;
	v[2] = e[0].f;
    	for (i = 3; i < N; i++)
		v[i] = 0;

	// Insercao da primeira aresta da lista de economias na matriz de adjacencias.
	ma[v[1] -1][v[2] -1] = ma[v[2] -1][v[1] -1] = 1;
	
	// Percorre a lista de economias.
	for (i = 1; i < n; i++)
	{
        	k = p = 0;
		// Percorre a matriz de adjacencias, para verificar se a insercao da aresta na posicao i da lista de
		// economias eh viavel.
	        for (j = 0; j < N; j++)
        	{
	        	k = k + ma[(e[i].i) - 1][j];
		        p = p + ma[(e[i].f) - 1][j];
	        }
        	if ((k + p) < 2)
        	{
			// Verifica se as cidades i e f ja estao no vetor.
            		k = verifica (v, e[i].i);
			p = verifica (v, e[i].f);
			// Se o valor de k for N, siginifica que a cidade i nao esta no vetor.
		    	if (k != N)
		    	{
				// Verifica se tem alguma cidade, no vetor, depois de i.
				if (v[k + 1] == 0)
					// Se naum houver, insere f no vetor, depois de i.
					insere_vetor (v, e[i].f, k + 1);
			    	else
					// Se houver, insere f no vetor, na posicao de i.
					insere_vetor (v, e[i].f, k);
				// Atualiza a matriz de adjacencias, indicando que existe uma aresta entre i e f.
				ma[(e[i].i) - 1][(e[i].f) - 1] = ma [(e[i].f) - 1][(e[i].i) - 1] = 1;
	    		}
			// Se o valor de p for N, siginifica que a cidade f nao esta no vetor.
		    	if (p != N)
		    	{
				// Verifica se tem alguma cidade, no vetor, depois de f.
				if (v[p + 1] == 0)
					// Se naum houver, insere i no vetor, depois de f.
					insere_vetor (v, e[i].i, p + 1);
			    	else
					// Se houver, insere i no vetor, na posicao de f.
					insere_vetor (v, e[i].i, p);
				// Atualiza a matriz de adjacencias, indicando que existe uma aresta entre i e f.
				ma[(e[i].i) - 1][(e[i].f) - 1] = ma [(e[i].i) - 1][(e[i].f) - 1] = 1;
	    		}
		}
		// Se o vetor jah estah completo, pare.
		if (v[N - 1] != 0)
			break;
	}

	// Apaga a lista, para desocupar a memoria.
	free (e);
}

int main ()
{
	char c[T], tipo;
	int i, v[N], **d;
	float x[N], y[N];
	
	// Alocacao de memoria para a matriz de distancias.
	d = (int **) malloc (N * sizeof (int *));
	if (!d)
	{
		printf("Memoria Insuficiente!\n"); // Mensagem de erro.
		exit (0);
	}
	for(i=0; i<N; i++)
	{
		d[i] = (int *) malloc (N * sizeof (int));
		if (!d[i])
		{
			printf("Memoria Insuficiente!\n"); // Mensagem de erro.
			exit (0);
		}
	}
	
	// Leitura do arquivo de entrada e do tipo do arquivo.
	printf ("\n\nDigite o nome do arquivo de entrada: "); // <nome do arquivo>.txt
	scanf ("%s", c);
	fflush (stdin);
	printf ("\n\nInforme o tipo do arquivo (ATT = a, EUC = e, GEO = g, Matriz = m): ");
	scanf (" %c", &tipo);

	// Verificacao do tipo do arquivo e calculo da distancia.
	switch (tipo)
	{
		case 'a': le_arq (x, y, c);
			  att (x, y, d);
			  break;
		case 'e': le_arq (x, y, c);
			  euc (x, y, d);
			  break;
		case 'g': le_arq (x, y, c);
			  geo (x, y, d);
			  break;
		case 'm': le_arq_matriz (d, c);
	}

	// Obtencao da solucao inicial.
	econ (v, d, 1);

	// Imprime a solucao inicial.
	for (i = 0; i < N; i++)
		printf ("\nv[%d] = %d", i, v[i]);

	return 0;
}
